#![feature(cursor_remaining)]

use std::fmt::Display;
use std::io::{Read, ErrorKind, Cursor, Write};
use std::net::{TcpListener, TcpStream, SocketAddr};
use std::thread;

use anyhow::{Result, bail};
use datatypes::lenght_prefixed_string::LenghtPrefixedString;
use itertools::Itertools;
use packets::serverbound::login::LoginStartPacket;
use packets::serverbound::status::StatusPingRequestPacket;
use packets::{Packet, Encode};
use packets::serverbound::handshake::HandshakePacket;
use packets::clientbound::status::{StatusResponsePacket, StatusPongResponsePacket};
use packets::clientbound::login::LoginSuccessPacket;
use serde::Serialize;
use thiserror::Error;
use uuid::Uuid;

use crate::datatypes::varint::VarInt;
use crate::packets::Decode;

pub mod datatypes;
pub mod packets;

fn main() {
    let listener = TcpListener::bind("127.0.0.1:25565").unwrap();
    println!("listening started, ready to accept");
    for stream in listener.incoming() {
        match stream {
            Err(e) => println!("Accept ERROR: {}", e),
            Ok(stream) => drop(thread::spawn(|| handle_client(stream))),
        }
    }
}

fn handle_client(stream: TcpStream) -> Result<()> {
    let mut client = Client::new(stream)?;

    loop {
        client.read_packet()?;
    }
}

#[derive(Debug, Clone, Copy)]
enum ConnectionState {
    Handshake,
    Status,
    Login,
    Play,
}

impl Display for ConnectionState {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ConnectionState::Handshake => f.write_str("Handshake"),
            ConnectionState::Status => f.write_str("Status"),
            ConnectionState::Login => f.write_str("Login"),
            ConnectionState::Play => f.write_str("Play"),
        }
    }
}

struct Client {
    state: ConnectionState,
    address: SocketAddr,
    stream: TcpStream,
}

impl Client {
    fn new(stream: TcpStream) -> Result<Client> {
        let address = stream.peer_addr()?;
        
        Ok(Client {
            state: ConnectionState::Handshake,
            address, stream,
        })
    }

    fn send_packet<T: Packet>(&mut self, packet: T) -> Result<()> {
        let mut data: Vec<u8> = Vec::new();
        // Encode packet ID
        VarInt {value: packet.get_id()}.encode(&mut data)?;
        // Encode packet data
        packet.encode(&mut data)?;
        
        // Prepend packet data
        let mut lenght: Vec<u8> = Vec::new();
        VarInt {value: data.len() as i32}.encode(&mut lenght)?;
        lenght.append(&mut data);
        // Data is empty now
        let data = lenght;

        self.stream.write_all(&data)?;
        Ok(())
    }
    
    fn read_packet(&mut self) -> Result<()> {
        let mut buffer = vec![0u8; 256];
        let read_bytes = self.stream.read(&mut buffer)?;
        if read_bytes == 0 {
            return Err(std::io::Error::new(ErrorKind::UnexpectedEof, "Read 0 bytes").into());
        }
        let bytes = &buffer[..read_bytes];
        let mut cursor = Cursor::new(bytes);

        let length: i32 = VarInt::decode(&mut cursor)?.into();
        let id: i32 = VarInt::decode(&mut cursor)?.into();


        self.handle_packet(id, length, cursor)
    }

    fn handle_packet(&mut self, id: i32, length: i32, cursor: Cursor<&[u8]>) -> Result<()> {
        match self.state {
            ConnectionState::Handshake => self.handle_handshake_packets(id, length, cursor),
            ConnectionState::Status => self.handle_status_packets(id, length, cursor),
            ConnectionState::Login => self.handle_login_packet(id, length, cursor),
            ConnectionState::Play => todo!(),
        }
    }

    fn handle_handshake_packets(&mut self, id: i32, length: i32, mut cursor: Cursor<&[u8]>) -> Result<()> {
        match id {
            0 => {
                let packet = HandshakePacket::decode(&mut cursor)?;
                println!("[HANDSHAKE] <- {}: Handshake request: IP: {}:{}", self.address, packet.server_address, packet.server_port);
                self.change_state(packet.next_state.into())
            },
            _ => {
                println!("[HANDSHAKE] <- {}: [ID: {}, Length: {}] {:02x}", self.address, id, length, cursor.remaining_slice().iter().format(" "));
                Ok(())
                //bail!(Error::InvalidPacketIdError {id, state: self.state}),
            }
        }
    }

    fn handle_status_packets(&mut self, id: i32, length: i32, mut cursor: Cursor<&[u8]>) -> Result<()> {
        match id {
            0 => {
                let response_data = ServerStatusResponseData {
                    version: ServerStatusResponseDataVersion { name: "Rust 1.19".to_owned(), protocol: 761 },
                    players: ServerStatusResponseDataPlayers { max: 100, online: 0 },
                    description: ServerStatusResponseDataDescription { text: "A rust server!".to_owned() },
                };

                let response = StatusResponsePacket {json: LenghtPrefixedString {value: serde_json::to_string(&response_data)?}};

                self.send_packet(response)?;
                
                Ok(())
            },
            1 => {
                let packet = StatusPingRequestPacket::decode(&mut cursor)?;
                let response = StatusPongResponsePacket {payload: packet.payload};

                self.send_packet(response)?;
                
                Ok(())
            },
            _ => {
                println!("[STATUS] <- {}: [ID: {}, Length: {}] {:02x}", self.address, id, length, cursor.remaining_slice().iter().format(" "));
                Ok(())
                //bail!(Error::InvalidPacketIdError {id, state: self.state}),
            }
        }

    }

    fn handle_login_packet(&mut self, id: i32, length: i32, mut cursor: Cursor<&[u8]>) -> Result<()> {
        match id {
            0 => {
                let packet = LoginStartPacket::decode(&mut cursor)?;
                let response = LoginSuccessPacket {
                    uuid: Uuid::from_u128(0xd15b7518f4d747f8b385f10bffedf643),
                    username: LenghtPrefixedString {value: "test".to_owned()},
                    property_length: VarInt {value: 0},
                };

                self.send_packet(response)?;
                self.change_state(3)?;

                Ok(())
            },
            _ => {
                println!("[LOGIN] <- {}: [ID: {}, Length: {}] {:02x}", self.address, id, length, cursor.remaining_slice().iter().format(" "));
                Ok(())
                //bail!(Error::InvalidPacketIdError {id, state: self.state}),
            }
        }
    }

    fn change_state(&mut self, state_index: i32) -> Result<()> {
        println!("<- {}: Moving connection to state: {}", self.address, state_index);
        match state_index {
            1 => self.state = ConnectionState::Status,
            2 => self.state = ConnectionState::Login,
            3 => self.state = ConnectionState::Play,
            _ => bail!(Error::InvalidNextState {next_state: state_index})
        }

        Ok(())
    }
}

#[derive(Error, Debug)]
enum Error {
    #[error("Packet has invalid packet id {id} for current state {state}")]
    InvalidPacketIdError {id: i32, state: ConnectionState},
    #[error("Requested to move to invalid next state {next_state}.")]
    InvalidNextState {next_state: i32},
}

#[derive(Debug, Serialize)]
struct ServerStatusResponseData {
    version: ServerStatusResponseDataVersion,
    players: ServerStatusResponseDataPlayers,
    description: ServerStatusResponseDataDescription,
}

#[derive(Debug, Serialize)]
struct ServerStatusResponseDataVersion {
    name: String,
    protocol: i32,
}

#[derive(Debug, Serialize)]
struct ServerStatusResponseDataPlayers {
    max: i32,
    online: i32,
}

#[derive(Debug, Serialize)]
struct ServerStatusResponseDataDescription {
    text: String,
}

